#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "../plugin/atanua_dll_interface.h"

static const char plugin_name1[]="2708";
static const char plugin_name2[]="2716";
static const char plugin_name3[]="2732";
static const int chipcount = 3;

/* Get the number of chips this DLL/interface supports */
ATANUADLLEXPORT void ATANUACALL getatanuadllinfo(atanuadllinfo *aDllInfo)
{
    /* DLL interface version. Set to 0. Set to -1 if in error. */
    aDllInfo->mDllVersion = 0;   
    /* Number of chips provided by this DLL */
    aDllInfo->mChipCount = chipcount;    
    /* Pointers to chip names. These may not point to stack! */
    aDllInfo->mChipName = new const char*[chipcount]; 
    /* Category numbers for chips. May be NULL. Categories are 0=dont care, 1=base, 2=chip, 3=I/O */
    aDllInfo->mChipCategory = new int[chipcount]; 

    aDllInfo->mChipName[0] = plugin_name1;
    aDllInfo->mChipCategory[0] = 2;
    aDllInfo->mChipName[1] = plugin_name2;
    aDllInfo->mChipCategory[1] = 2;
    aDllInfo->mChipName[2] = plugin_name3;
    aDllInfo->mChipCategory[2] = 2;
}

/* create a chip */
ATANUADLLEXPORT int ATANUACALL create(chipinfostruct *aChipInfo, const char *aChipname)
{
    int whichchip = 0;
    if (strcmp(aChipname, plugin_name1) == 0)
        whichchip = 1;
    if (strcmp(aChipname, plugin_name2) == 0)
        whichchip = 2;
    if (strcmp(aChipname, plugin_name3) == 0)
        whichchip = 3;

    if (whichchip == 0)
        return 0;

    int pincount = 0;
    switch (whichchip)
    {
    case 1:
        pincount = 18; // 8 output, 10 address
        break;
    case 2:
        pincount = 19; // 8 output, 11 address
        break;
    case 3:
        pincount = 20; // 8 output, 12 address
        break;
    }

    memset(aChipInfo, 0, sizeof(chipinfostruct));

    /* Number of pins for the structure */
    aChipInfo->mPinCount = pincount;
    /* Pin input data states. See netstates enum. */
    aChipInfo->mPinInput = new int[pincount];    

    /* Pin output data states. See pinmodes enum. */
    aChipInfo->mPinOutput = new int[pincount];

    int i;
    for (i = 0; i < pincount; i++)
        aChipInfo->mPinOutput[i] = PINSTATE_READ;

    aChipInfo->mWidth = 6.8f;      /* Chip width */
    aChipInfo->mHeight = 3.4f;     /* Chip height */

    /* Pin coordinates in relation to the chip. Two floats per pin. */
    aChipInfo->mPinCoordinates = new float[pincount*2]; 

    float xpos = 0.13f;
    int p = 0;

    xpos += 0.5475f; // skip VCC

    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = -0.25; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = -0.25; p++;

    if (whichchip == 3)
    {
        aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
        aChipInfo->mPinCoordinates[p*2+1] = -0.25; p++;
    }
    else
    {
        xpos += 0.5475f; // skip VBB
    }

    xpos += 0.5475f; // skip CS(PE)
    
    if (whichchip != 1)
    {
        aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
        aChipInfo->mPinCoordinates[p*2+1] = -0.25; p++;
    }
    else
    {
        xpos += 0.5475f; // skip VDD
    }

    xpos += 0.5475f; // skip program

    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = -0.25f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = -0.25f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = -0.25f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = -0.25f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = -0.25f; p++;

    xpos = 0.13f;

    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;
    aChipInfo->mPinCoordinates[p*2+0] = xpos; xpos += 0.5475f;
    aChipInfo->mPinCoordinates[p*2+1] = 3.1f; p++;

    /* Texture filename. Can be NULL. */
    aChipInfo->mTextureFilename = "data/chip_24pin.png"; 

    switch (whichchip)
    {
    case 1:
        /* Chip name rendered at the center of the chip */
        aChipInfo->mChipName = "2708";        

        /* Chip tooltip. Can be NULL. */
        aChipInfo->mTooltip = "1024-word by 8-bit\n"
                              "erasable programmabe\n"
                              "read only memories\n\n"
                              "Click chip to give it\n"
                              "keyboard focus, then\n"
                              "press L to load ROM\n"
                              "contents from 2708.bin";
        break;
    case 2:
        /* Chip name rendered at the center of the chip */
        aChipInfo->mChipName = "2716";        

        /* Chip tooltip. Can be NULL. */
        aChipInfo->mTooltip = "2048-word by 8-bit\n"
                              "erasable programmabe\n"
                              "read only memories\n\n"
                              "Click chip to give it\n"
                              "keyboard focus, then\n"
                              "press L to load ROM\n"
                              "contents from 2716.bin";
        break;
    case 3:
        /* Chip name rendered at the center of the chip */
        aChipInfo->mChipName = "2732";        

        /* Chip tooltip. Can be NULL. */
        aChipInfo->mTooltip = "4096-word by 8-bit\n"
                              "erasable programmabe\n"
                              "read only memories\n\n"
                              "Click chip to give it\n"
                              "keyboard focus, then\n"
                              "press L to load ROM\n"
                              "contents from 2732.bin";
        break;
    }

    /* Tooltips for the pins. Can be NULL. */
    p = 0;
    aChipInfo->mPinTooltips = new const char*[pincount];
    aChipInfo->mPinTooltips[p] = "A8"; p++;
    aChipInfo->mPinTooltips[p] = "A9"; p++;
    if (whichchip == 3)
    {
      aChipInfo->mPinTooltips[p] = "A11"; p++;
    }
    if (whichchip != 1)
    {
      aChipInfo->mPinTooltips[p] = "A10"; p++;
    }
    aChipInfo->mPinTooltips[p] = "Q7"; p++;
    aChipInfo->mPinTooltips[p] = "Q6"; p++;
    aChipInfo->mPinTooltips[p] = "Q5"; p++;
    aChipInfo->mPinTooltips[p] = "Q4"; p++;
    aChipInfo->mPinTooltips[p] = "Q3"; p++;
    aChipInfo->mPinTooltips[p] = "A7"; p++;
    aChipInfo->mPinTooltips[p] = "A6"; p++;
    aChipInfo->mPinTooltips[p] = "A5"; p++;
    aChipInfo->mPinTooltips[p] = "A4"; p++;
    aChipInfo->mPinTooltips[p] = "A3"; p++;
    aChipInfo->mPinTooltips[p] = "A2"; p++;
    aChipInfo->mPinTooltips[p] = "A1"; p++;
    aChipInfo->mPinTooltips[p] = "A0"; p++;
    aChipInfo->mPinTooltips[p] = "Q0"; p++;
    aChipInfo->mPinTooltips[p] = "Q1"; p++;
    aChipInfo->mPinTooltips[p] = "Q2"; 

    switch (whichchip)
    {
    case 1:
        /* Number of bytes in persistent data store. Can be 0. */
        aChipInfo->mPersistentDataCount = 1024; 

        /* Pointer to the persistent data store. Can be NULL. */
        aChipInfo->mPersistentDataPtr = new char[1024]; 

        /* User pointer for storing other chip-related data. */
        aChipInfo->mUserPtr = (void*)1;
        break;
    case 2:
        /* Number of bytes in persistent data store. Can be 0. */
        aChipInfo->mPersistentDataCount = 2048; 

        /* Pointer to the persistent data store. Can be NULL. */
        aChipInfo->mPersistentDataPtr = new char[2048]; 

        /* User pointer for storing other chip-related data. */
        aChipInfo->mUserPtr = (void*)2;
        break;
    case 3:
        /* Number of bytes in persistent data store. Can be 0. */
        aChipInfo->mPersistentDataCount = 4096; 

        /* Pointer to the persistent data store. Can be NULL. */
        aChipInfo->mPersistentDataPtr = new char[4096]; 

        /* User pointer for storing other chip-related data. */
        aChipInfo->mUserPtr = (void*)3;
        break;
    }

    return 1;
}

/* Update a chip */
ATANUADLLEXPORT void ATANUACALL update(chipinfostruct *aChipInfo, float aTick)
{
    int whichchip = (int)aChipInfo->mUserPtr;
    int pos;
    int p = 0;
    switch (whichchip)
    {
    case 1:
            pos = ((aChipInfo->mPinInput[14] == NETSTATE_HIGH)?(1 << 0):0) |
                  ((aChipInfo->mPinInput[13] == NETSTATE_HIGH)?(1 << 1):0) |
                  ((aChipInfo->mPinInput[12] == NETSTATE_HIGH)?(1 << 2):0) |
                  ((aChipInfo->mPinInput[11] == NETSTATE_HIGH)?(1 << 3):0) |
                  ((aChipInfo->mPinInput[10] == NETSTATE_HIGH)?(1 << 4):0) |
                  ((aChipInfo->mPinInput[9] == NETSTATE_HIGH)?(1 << 5):0) |
                  ((aChipInfo->mPinInput[8] == NETSTATE_HIGH)?(1 << 6):0) |
                  ((aChipInfo->mPinInput[7] == NETSTATE_HIGH)?(1 << 7):0) |
                  ((aChipInfo->mPinInput[0] == NETSTATE_HIGH)?(1 << 8):0) |
                  ((aChipInfo->mPinInput[1] == NETSTATE_HIGH)?(1 << 9):0);
            p = 2;
            break;
    case 2:
            pos = ((aChipInfo->mPinInput[15] == NETSTATE_HIGH)?(1 << 0):0) |
                  ((aChipInfo->mPinInput[14] == NETSTATE_HIGH)?(1 << 1):0) |
                  ((aChipInfo->mPinInput[13] == NETSTATE_HIGH)?(1 << 2):0) |
                  ((aChipInfo->mPinInput[12] == NETSTATE_HIGH)?(1 << 3):0) |
                  ((aChipInfo->mPinInput[11] == NETSTATE_HIGH)?(1 << 4):0) |
                  ((aChipInfo->mPinInput[10] == NETSTATE_HIGH)?(1 << 5):0) |
                  ((aChipInfo->mPinInput[9] == NETSTATE_HIGH)?(1 << 6):0) |
                  ((aChipInfo->mPinInput[8] == NETSTATE_HIGH)?(1 << 7):0) |
                  ((aChipInfo->mPinInput[0] == NETSTATE_HIGH)?(1 << 8):0) |
                  ((aChipInfo->mPinInput[1] == NETSTATE_HIGH)?(1 << 9):0) |
                  ((aChipInfo->mPinInput[2] == NETSTATE_HIGH)?(1 << 10):0);
            p = 3;
            break;
    case 3:
            pos = ((aChipInfo->mPinInput[16] == NETSTATE_HIGH)?(1 << 0):0) |
                  ((aChipInfo->mPinInput[15] == NETSTATE_HIGH)?(1 << 1):0) |
                  ((aChipInfo->mPinInput[14] == NETSTATE_HIGH)?(1 << 2):0) |
                  ((aChipInfo->mPinInput[13] == NETSTATE_HIGH)?(1 << 3):0) |
                  ((aChipInfo->mPinInput[12] == NETSTATE_HIGH)?(1 << 4):0) |
                  ((aChipInfo->mPinInput[11] == NETSTATE_HIGH)?(1 << 5):0) |
                  ((aChipInfo->mPinInput[10] == NETSTATE_HIGH)?(1 << 6):0) |
                  ((aChipInfo->mPinInput[9] == NETSTATE_HIGH)?(1 << 7):0) |
                  ((aChipInfo->mPinInput[0] == NETSTATE_HIGH)?(1 << 8):0) |
                  ((aChipInfo->mPinInput[1] == NETSTATE_HIGH)?(1 << 9):0) |
                  ((aChipInfo->mPinInput[2] == NETSTATE_HIGH)?(1 << 10):0) |
                  ((aChipInfo->mPinInput[3] == NETSTATE_HIGH)?(1 << 11):0);
            p = 4;
            break;
    }
    int data = (unsigned char)aChipInfo->mPersistentDataPtr[pos];

    aChipInfo->mPinOutput[p] = (data & (1 << 7)) ? PINSTATE_WRITE_HIGH : PINSTATE_WRITE_LOW; p++;
    aChipInfo->mPinOutput[p] = (data & (1 << 6)) ? PINSTATE_WRITE_HIGH : PINSTATE_WRITE_LOW; p++;
    aChipInfo->mPinOutput[p] = (data & (1 << 5)) ? PINSTATE_WRITE_HIGH : PINSTATE_WRITE_LOW; p++;
    aChipInfo->mPinOutput[p] = (data & (1 << 4)) ? PINSTATE_WRITE_HIGH : PINSTATE_WRITE_LOW; p++;
    aChipInfo->mPinOutput[p] = (data & (1 << 3)) ? PINSTATE_WRITE_HIGH : PINSTATE_WRITE_LOW; p++;
    p += 8;
    aChipInfo->mPinOutput[p] = (data & (1 << 0)) ? PINSTATE_WRITE_HIGH : PINSTATE_WRITE_LOW; p++;
    aChipInfo->mPinOutput[p] = (data & (1 << 1)) ? PINSTATE_WRITE_HIGH : PINSTATE_WRITE_LOW; p++;
    aChipInfo->mPinOutput[p] = (data & (1 << 2)) ? PINSTATE_WRITE_HIGH : PINSTATE_WRITE_LOW; p++;
}

/* Render and/or handle key input. Return 0 to let Atanua render the chip. */
ATANUADLLEXPORT int ATANUACALL render(chipinfostruct *aChipInfo, int aKey)
{
    int whichchip = (int)aChipInfo->mUserPtr;
    if (aKey == 'l')
    {
        FILE * f;
        switch (whichchip)
        {
        case 1:
            f = fopen("2708.bin", "rb");
            break;
        case 2:
            f = fopen("2716.bin", "rb");
            break;
        case 3:
            f = fopen("2732.bin", "rb");
            break;
        }
        if (f)
        {
            fread(aChipInfo->mPersistentDataPtr, aChipInfo->mPersistentDataCount, 1, f);
            fclose(f);
        }
    }
    // Let Atanua handle rendering
    return 0;
}

/* Called when chip is being destroyed. */
ATANUADLLEXPORT void ATANUACALL cleanup(chipinfostruct *aChipInfo)
{   
    // Free the bits that we allocated. 
    // Atanua has no idea how we allocated these - could be globals, for instance.
    delete[] aChipInfo->mPinOutput;   
    delete[] aChipInfo->mPinInput;    
    delete[] aChipInfo->mPinCoordinates; 
    delete[] aChipInfo->mPinTooltips;
    delete[] aChipInfo->mPersistentDataPtr;
    return;
}
